/*
* Author: Chris Campbell - www.iforce2d.net
*
* Copyright (c) 2006-2011 Erin Catto http://www.box2d.org
*
* This software is provided 'as-is', without any express or implied
* warranty.  In no event will the authors be held liable for any damages
* arising from the use of this software.
* Permission is granted to anyone to use this software for any purpose,
* including commercial applications, and to alter it and redistribute it
* freely, subject to the following restrictions:
* 1. The origin of this software must not be misrepresented; you must not
* claim that you wrote the original software. If you use this software
* in a product, an acknowledgment in the product documentation would be
* appreciated but is not required.
* 2. Altered source versions must be plainly marked as such, and must not be
* misrepresented as being the original software.
* 3. This notice may not be removed or altered from any source distribution.
*/

#ifndef IFORCE2D_EXPLOSIONS_H
#define IFORCE2D_EXPLOSIONS_H

#ifndef DEGTORAD
#define DEGTORAD 0.0174532925199432957f
#define RADTODEG 57.295779513082320876f
#endif

#include <vector>

enum _explosionType {
    ET_PROXIMITY,
    ET_RAYCAST,
    ET_PARTICLE,

    ET_MAX
};

#define MAX_BLAST_RAYS 128

//keep these in global scope so resetting test does not affect it
_explosionType g_explosionType = ET_PROXIMITY;
int g_numRays = 32;
bool g_pauseAfterBlast = false;

//subclass b2QueryCallback for proximity query callback
class MyQueryCallback : public b2QueryCallback {
public:
    std::vector<b2Body*> foundBodies;

    bool ReportFixture(b2Fixture* fixture) {
        foundBodies.push_back( fixture->GetBody() );
        return true;//keep going to find all fixtures in the query area
    }
};

//simple ray cast callback to find the closest body the ray hit
class RayCastClosestCallback : public b2RayCastCallback
{
public:
    b2Body* m_body;
    b2Vec2 m_point;

    RayCastClosestCallback() { m_body = NULL; }

    float32 ReportFixture(b2Fixture* fixture, const b2Vec2& point, const b2Vec2& normal, float32 fraction)
    {
        m_body = fixture->GetBody();
        m_point = point;
        return fraction;
    }
};

class iforce2d_Explosions: public Test
{
public:
    iforce2d_Explosions()
    {
        //include dumped world data
        #include "explosions.cpp"

        //find the only body in the world which is a circle of radius 0.25
        m_grenadeBody = NULL;
        for (b2Body* b = m_world->GetBodyList(); b; b = b->GetNext()) {
            if ( !b->GetFixtureList() )
                continue;
            b2Shape* shape = b->GetFixtureList()->GetShape();
            if ( shape->GetType() == b2Shape::e_circle ) {
                if ( ((b2CircleShape*)shape)->m_radius < 0.3f ) {
                    m_grenadeBody = b;
                    break;
                }
            }
        }

        m_blastRadius = 10;
        m_blastPower = 1000;

        memset(m_blastParticleBodies, 0, sizeof(m_blastParticleBodies));
        m_shouldPause = false;

        glEnable(GL_POINT_SMOOTH);
    }

    //this is the same for proximity and raycast methods so we can put it in a common function
    void applyBlastImpulse(b2Body* body, b2Vec2 blastCenter, b2Vec2 applyPoint, float blastPower)
    {
        //ignore the grenade itself, and any non-dynamic bodies
        if ( body == m_grenadeBody || body->GetType() != b2_dynamicBody )
            return;
        b2Vec2 blastDir = applyPoint - blastCenter;
        float distance = blastDir.Normalize();
        //ignore bodies exactly at the blast point - blast direction is undefined
        if ( distance == 0 )
            return;
        float invDistance = 1 / distance;
        float impulseMag = blastPower * invDistance * invDistance;
        impulseMag = b2Min( impulseMag, 500.0f );
        body->ApplyLinearImpulse( impulseMag * blastDir, applyPoint );
    }

    void display_proximity()
    {
        b2Vec2 center = m_grenadeBody->GetPosition();

        //find all fixtures within blast radius AABB
        MyQueryCallback queryCallback;
        b2AABB aabb;
        aabb.lowerBound = center - b2Vec2( m_blastRadius, m_blastRadius );
        aabb.upperBound = center + b2Vec2( m_blastRadius, m_blastRadius );
        m_world->QueryAABB( &queryCallback, aabb );

        //check which of these have their center of mass within the blast radius
        glEnable(GL_BLEND);
        glPointSize(8);
        for (int i = 0; i < queryCallback.foundBodies.size(); i++) {
            b2Body* body = queryCallback.foundBodies[i];
            if ( body == m_grenadeBody || body->GetType() != b2_dynamicBody )
                continue;
            b2Vec2 bodyCom = body->GetWorldCenter();
            if ( (bodyCom - center).Length() > m_blastRadius )
                continue;
            //draw line and dot to show which bodies will get blasted
            glColor4f(1,1,0,0.5);
            glBegin(GL_LINES);
            glVertex2fv( (GLfloat*)&center );
            glVertex2fv( (GLfloat*)&bodyCom );
            glEnd();
            glBegin(GL_POINTS);
            glVertex2fv( (GLfloat*)&bodyCom );
            glEnd();
        }
        glPointSize(1);

        //draw a circle to show the blast radius
        glColor4f(0.5f, 0.5f, 0.5f, 0.5f);
        glBegin(GL_LINE_LOOP);
        for (int i = 0; i < 32; i++) {
            float angle = (i / 32.0f) * 360 * DEGTORAD;
            b2Vec2 rayDir( sinf(angle), cosf(angle) );
            b2Vec2 rayEnd = center + m_blastRadius * rayDir;
            glVertex2fv( (GLfloat*)&rayEnd );
        }
        glEnd();

    }

    void explode_proximity()
    {
        b2Vec2 center = m_grenadeBody->GetPosition();

        //find all fixtures within blast radius AABB
        MyQueryCallback queryCallback;
        b2AABB aabb;
        aabb.lowerBound = center - b2Vec2( m_blastRadius, m_blastRadius );
        aabb.upperBound = center + b2Vec2( m_blastRadius, m_blastRadius );
        m_world->QueryAABB( &queryCallback, aabb );

        //check which of these have their center of mass within the blast radius
        for (int i = 0; i < queryCallback.foundBodies.size(); i++) {
            b2Body* body = queryCallback.foundBodies[i];
            b2Vec2 bodyCom = body->GetWorldCenter();
            //ignore bodies outside the blast range
            if ( (bodyCom - center).Length() >= m_blastRadius )
                continue;
            applyBlastImpulse(body, center, bodyCom, m_blastPower * 0.05f );//scale blast power to roughly match results of other methods at 32 rays
        }
    }

    void display_raycast(bool showRayHits)
    {
        glEnable(GL_BLEND);
        glPointSize(8);
        if ( !showRayHits ) {
            //just to visually differentiate this from the raycast mode
            glEnable(GL_LINE_STIPPLE);
            glLineStipple( 1, 0xF0F0 );
        }
        if ( m_grenadeBody ) {
            //draw all rays around grenade
            b2Vec2 center = m_grenadeBody->GetPosition();
            for (int i = 0; i < g_numRays; i++) {
                float angle = (i / (float)g_numRays) * 360 * DEGTORAD;
                b2Vec2 rayDir( sinf(angle), cosf(angle) );
                b2Vec2 rayEnd = center + m_blastRadius * rayDir;
                glColor4f(0.5f,0.5f,0.5f,0.5f);
                glLineWidth(1);
                glBegin(GL_LINES);
                glVertex2fv( (GLfloat*)&rayEnd );
                glVertex2fv( (GLfloat*)&center );
                glEnd();

                if ( showRayHits ) {
                    //check what this ray hits
                    RayCastClosestCallback callback;
                    m_world->RayCast(&callback, center, rayEnd);
                    if ( callback.m_body ) {
                        //draw line and dot to show which bodies will get blasted
                        glColor4f(1,1,0,0.5);
                        glLineWidth(2);
                        glBegin(GL_LINES);
                        glVertex2fv( (GLfloat*)&center );
                        glVertex2fv( (GLfloat*)&callback.m_point );
                        glEnd();
                        glBegin(GL_POINTS);
                        glVertex2fv( (GLfloat*)&callback.m_point );
                        glEnd();
                    }
                }
            }
        }
        glLineWidth(1);
        glPointSize(1);
        glDisable(GL_LINE_STIPPLE);
    }

    void explode_raycast()
    {
        if ( m_grenadeBody ) {
            b2Vec2 center = m_grenadeBody->GetPosition();
            for (int i = 0; i < g_numRays; i++) {
                float angle = (i / (float)g_numRays) * 360 * DEGTORAD;
                b2Vec2 rayDir( sinf(angle), cosf(angle) );
                b2Vec2 rayEnd = center + m_blastRadius * rayDir;

                //check what this ray hits
                RayCastClosestCallback callback;
                m_world->RayCast(&callback, center, rayEnd);
                if ( callback.m_body ) {
                    applyBlastImpulse(callback.m_body, center, callback.m_point, (m_blastPower / (float)g_numRays));
                }
            }
        }
    }

    void display_particle()
    {
        //dashed lines to show where particles will go
        //display_raycast(false);

        //particle previous position trail
        glEnable(GL_BLEND);
        if ( !m_previousParticlePositions.empty() ) {
            for (int i = 0; i < MAX_BLAST_RAYS; i++) {
                if ( m_blastParticleBodies[i] ) {
                    glBegin(GL_LINES);
                    for (int k = 1; k < m_previousParticlePositions.size(); k++) {
                        float alpha = (k-1) / (float)m_previousParticlePositions.size();
                        glColor4f(1,1,0, alpha);
                        glVertex2fv( (GLfloat*)&(m_previousParticlePositions[k-1][i]) );
                        alpha = k / (float)m_previousParticlePositions.size();
                        glColor4f(1,1,0, alpha);
                        glVertex2fv( (GLfloat*)&(m_previousParticlePositions[k][i]) );
                    }
                    int k = m_previousParticlePositions.size()-1;
                    float alpha = k / (float)m_previousParticlePositions.size();
                    glColor4f(1,1,0, alpha);
                    glVertex2fv( (GLfloat*)&(m_previousParticlePositions[k][i]) );
                    glColor4f(1,1,0,1);
                    b2Vec2 currentPos = m_blastParticleBodies[i]->GetPosition();
                    glVertex2fv( (GLfloat*)&currentPos );
                    glEnd();
                }
            }
        }
    }

    void explode_particle()
    {
        //clear old particles
        for (int i = 0; i < MAX_BLAST_RAYS; i++) {
            if ( m_blastParticleBodies[i] ) {
                m_world->DestroyBody(m_blastParticleBodies[i]);
                m_blastParticleBodies[i] = NULL;
            }
        }

        //clear previous positions
        for (int k = 0; k < m_previousParticlePositions.size(); k++)
            delete[] m_previousParticlePositions[k];
        m_previousParticlePositions.clear();

        if ( m_grenadeBody ) {
            b2Vec2 center = m_grenadeBody->GetPosition();
            for (int i = 0; i < g_numRays; i++) {
                float angle = (i / (float)g_numRays) * 360 * DEGTORAD;
                b2Vec2 rayDir( sinf(angle), cosf(angle) );

                b2BodyDef bd;
                bd.type = b2_dynamicBody;
                bd.fixedRotation = true;
                bd.bullet = true;
                bd.linearDamping = 10;
                bd.gravityScale = 0;
                bd.position = center;
                bd.linearVelocity = 0.125f * m_blastPower * rayDir;
                b2Body* body = m_world->CreateBody(&bd);

                b2CircleShape circleShape;
                circleShape.m_radius = 0.05;

                b2FixtureDef fd;
                fd.shape = &circleShape;
                fd.density = 60 / (float)g_numRays;
                fd.friction = 0;
                fd.restitution = 0.99f;
                fd.filter.groupIndex = -1;
                body->CreateFixture(&fd);

                m_blastParticleBodies[i] = body;
            }
        }
    }

    void Keyboard(unsigned char key)
    {
        switch (key) {
        case 'a' : g_explosionType = (_explosionType)((g_explosionType+1) % ET_MAX); break;
        case 'b' :
        {
            switch ( g_explosionType ) {
            case ET_PROXIMITY: explode_proximity(); m_shouldPause = g_pauseAfterBlast; break;
            case ET_RAYCAST: explode_raycast(); m_shouldPause = g_pauseAfterBlast; break;
            case ET_PARTICLE: explode_particle(); m_shouldPause = g_pauseAfterBlast; break;
            default:;
            }
        }
            break;
        case 's': g_numRays--; g_numRays = b2Clamp( g_numRays, 8, MAX_BLAST_RAYS ); break;
        case 'w': g_numRays++; g_numRays = b2Clamp( g_numRays, 8, MAX_BLAST_RAYS ); break;
        case 'q': g_pauseAfterBlast = !g_pauseAfterBlast;
        default: Test::Keyboard(key);
        }
    }

    void Step(Settings* settings)
    {
        m_debugDraw.DrawString(5, m_textLine, "Press a to change explosion type (current type: %s)",
                                                g_explosionType == ET_PROXIMITY ? "proximity" :
                                                g_explosionType == ET_RAYCAST ? "raycast" : "particle" );
        m_textLine += 15;
        m_debugDraw.DrawString(5, m_textLine, "Press b to go bang");
        m_textLine += 15;
        m_debugDraw.DrawString(5, m_textLine, "Press s/w to change the number of rays (current number: %d)", g_numRays);
        m_textLine += 15;
        m_debugDraw.DrawString(5, m_textLine, "Press q to toggle pausing after blast (currently: %s)", g_pauseAfterBlast ? "true" : "false");
        m_textLine += 15;

        if ( m_shouldPause ) {
            settings->pause = true;
            m_shouldPause = false;
        }

        //record positions of particles before stepping
        if ( g_explosionType == ET_PARTICLE && m_previousParticlePositions.size() < 20 && !settings->pause || settings->singleStep ) {
            b2Vec2* prevPositions = new b2Vec2[MAX_BLAST_RAYS];
            memset(prevPositions, 0, MAX_BLAST_RAYS * sizeof(b2Vec2));
            for (int i = 0; i < MAX_BLAST_RAYS; i++) {
                if ( m_blastParticleBodies[i] )
                    prevPositions[i] = m_blastParticleBodies[i]->GetPosition();
            }
            m_previousParticlePositions.push_back(prevPositions);
        }

        Test::Step(settings);

        switch ( g_explosionType ) {
        case ET_PROXIMITY: display_proximity(); break;
        case ET_RAYCAST: display_raycast(true); break;
        case ET_PARTICLE: display_particle(); break;
        default:;
        }
    }

    static Test* Create()
    {
        return new iforce2d_Explosions;
    }

    b2Body* m_grenadeBody;
    bool m_shouldPause;

    float m_blastRadius;
    float m_blastPower;

    b2Body* m_blastParticleBodies[MAX_BLAST_RAYS];

    std::vector<b2Vec2*> m_previousParticlePositions;
};

#endif









